/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/*
 *    ClusteredMetaInstanceClassifier.java
 *    Copyright (C) 2007 University of Waikato, Hamilton, New Zealand
 *
 */

package weka.classifiers.meta;

import weka.classifiers.RandomizableSingleClassifierEnhancer;
import weka.classifiers.rules.ZeroR;
import weka.clusterers.Clusterer;
import weka.clusterers.ClustersRetrievable;
import weka.clusterers.FirstKRelativeSquaredEuclideanDistanceClusterer;
import weka.clusterers.NumberOfClustersRequestable;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Option;
import weka.core.OptionHandler;
import weka.core.RevisionUtils;
import weka.core.TechnicalInformation;
import weka.core.TechnicalInformationHandler;
import weka.core.Utils;
import weka.core.TechnicalInformation.Field;
import weka.core.TechnicalInformation.Type;
import weka.filters.Filter;
import weka.filters.unsupervised.attribute.NominalToBinary;
import weka.filters.unsupervised.attribute.Normalize;
import weka.filters.unsupervised.attribute.ReplaceMissingValues;
import weka.filters.unsupervised.attribute.Standardize;
import weka.filters.unsupervised.instance.IdenticalInstanceFilter;

import java.util.Enumeration;
import java.util.Vector;

/**
 <!-- globalinfo-start -->
 * Class that implements a meta clustering classifier.<br/>
 * <br/>
 * For more information see:<br/>
 * <br/>
 * Reuben Evans (2008). Clustering for Classification: Using Standard Clustering Methods to Summarise Datasets with Minimal Loss of Classification Accuracy. VDM Verlag.
 * <p/>
 <!-- globalinfo-end -->
 * 
 <!-- technical-bibtex-start -->
 * BibTeX:
 * <pre>
 * &#64;book{Evans2008,
 *    author = {Reuben Evans},
 *    pages = {108},
 *    publisher = {VDM Verlag},
 *    title = {Clustering for Classification: Using Standard Clustering Methods to Summarise Datasets with Minimal Loss of Classification Accuracy},
 *    year = {2008},
 *    ISBN = {3639031636}
 * }
 * </pre>
 * <p/>
 <!-- technical-bibtex-end -->
 *
 <!-- options-start -->
 * Valid options are: <p/>
 * 
 * <pre> -C &lt;number&gt;
 *  Set the number of clusters to generate.
 *  (default = 0, i.e. sqrt(numExamples)).</pre>
 * 
 * <pre> -K &lt;classname+options&gt;
 *  Full name of base clusterer and its options.
 *  (default: Internal Method)</pre>
 * 
 * <pre> -I
 *  Enable weight equalisation.
 *  (default: off)</pre>
 * 
 * <pre> -T
 *  Enable distribution clustering.
 *  (default: off)</pre>
 * 
 * <pre> -H
 *  Threshold for cluster sizes.
 *  (default: 0.1)</pre>
 * 
 * <pre> -S &lt;num&gt;
 *  Random number seed.
 *  (default 1)</pre>
 * 
 * <pre> -D
 *  If set, classifier is run in debug mode and
 *  may output additional info to the console</pre>
 * 
 * <pre> -W
 *  Full name of base classifier.
 *  (default: weka.classifiers.rules.ZeroR)</pre>
 * 
 * <pre> 
 * Options specific to classifier weka.classifiers.rules.ZeroR:
 * </pre>
 * 
 * <pre> -D
 *  If set, classifier is run in debug mode and
 *  may output additional info to the console</pre>
 * 
 <!-- options-end -->
 * 
 * @author Bernhard Pfahringer, Reuben Evans
 * @version $Revision$
 */
public class ClusteredMetaInstanceClassifier
  extends RandomizableSingleClassifierEnhancer
  implements TechnicalInformationHandler {

  /** for serialization. */
  private static final long serialVersionUID = 200703021026L;

  /** Filter used for normalizing the data. */
  protected Standardize m_standardize;

  /** Filter used for normalizing the data. */
  protected Normalize m_normalize;

  /** Filter used for converting nominal values, because we can't handle them 
   * in the distance function. */
  protected NominalToBinary m_nominaltobinary;

  /** Filter used to replace missing vlaues, because we can't handle them. */
  protected ReplaceMissingValues m_missing;

  /** The number of clusters. */
  protected int m_numClusters = 0;

  /** the clusterer to use. */
  protected Clusterer m_Clusterer = null;

  /** Whether to take cluster distribution or only cluster number into account. */
  protected boolean m_Distribution = false;

  /** ??? */
  protected boolean m_EqualWeights = false;

  /** ??? */
  protected double m_Threshold = 0.0;

  /** ??? */
  protected IdenticalInstanceFilter m_Ident;

  /** the class index. */
  protected int m_classIndex = -1;
  
  /** the header information this also staore all the clusters when the 
   * classifier is built. */
  protected Instances m_header;
  
  /** the 0R model, in case we cannot train. */
  protected ZeroR m_ZeroR = null;

  /*
   * Assoicated with numeric class prediction this variable holds the 
   * difference between 0 and the standardised class value
   * Used to correct the predictions for numberic values to take into 
   * accounts the effects of standardisation/normalisation.
   */
  protected double m_Alin;
  protected double m_Blin;

  /** Variable that record when we are working with a numeric class value. */
  protected boolean m_numericClass;

  /**
   * default constructor.
   */
  public ClusteredMetaInstanceClassifier() {
    super();
    
    setClassifier(new weka.classifiers.rules.ZeroR()); // indicate "empty"...
    setClusterer(new FirstKRelativeSquaredEuclideanDistanceClusterer());
  }

  /**
   * Returns a string describing this classifier.
   * 
   * @return 		a description of the classifier suitable for
   * 			displaying in the explorer/experimenter gui
   */
  public String globalInfo() {
    return 
        "Class that implements a meta clustering classifier.\n\n"
      + "For more information see:\n\n"
      + getTechnicalInformation().toString();
  }

  /**
   * Returns an instance of a TechnicalInformation object, containing 
   * detailed information about the technical background of this class,
   * e.g., paper reference or book this class is based on.
   * 
   * @return 		the technical information about this class
   */
  public TechnicalInformation getTechnicalInformation() {
    TechnicalInformation 	result;
    
    result = new TechnicalInformation(Type.BOOK);
    result.setValue(Field.AUTHOR, "Reuben Evans");
    result.setValue(Field.YEAR, "2008");
    result.setValue(Field.TITLE, "Clustering for Classification: Using Standard Clustering Methods to Summarise Datasets with Minimal Loss of Classification Accuracy");
    result.setValue(Field.ISBN, "3639031636");
    result.setValue(Field.PAGES, "108");
    result.setValue(Field.PUBLISHER, "VDM Verlag");
    
    return result;
  }

  /**
   * Returns an enumeration describing the available options.
   *
   * @return 		an enumeration of all the available options
   */
  public Enumeration listOptions() {
    Vector result = new Vector();

    result.addElement(new Option(
	  "\tSet the number of clusters to generate.\n"
	  + "\t(default = 0, i.e. sqrt(numExamples)).",
	  "C", 1, "-C <number>"));

    result.addElement(new Option(
	  "\tFull name of base clusterer and its options.\n"
	  + "\t(default: Internal Method)",
	  "K", 1, "-K <classname+options>"));

    result.addElement(new Option(
	  "\tEnable weight equalisation.\n"
	  + "\t(default: off)",
	  "I", 0, "-I"));

    result.addElement(new Option(
	  "\tEnable distribution clustering.\n"
	  + "\t(default: off)",
	  "T", 0, "-T"));

    result.addElement(new Option(
	  "\tThreshold for cluster sizes.\n"
	  + "\t(default: 0.1)",
	  "H", 1, "-H"));

    Enumeration enu = super.listOptions();
    while (enu.hasMoreElements())
      result.addElement(enu.nextElement());

    return result.elements();
  }

  /**
   * Parses the options for this object. <p/>
   *
   <!-- options-start -->
   * Valid options are: <p/>
   * 
   * <pre> -C &lt;number&gt;
   *  Set the number of clusters to generate.
   *  (default = 0, i.e. sqrt(numExamples)).</pre>
   * 
   * <pre> -K &lt;classname+options&gt;
   *  Full name of base clusterer and its options.
   *  (default: Internal Method)</pre>
   * 
   * <pre> -I
   *  Enable weight equalisation.
   *  (default: off)</pre>
   * 
   * <pre> -T
   *  Enable distribution clustering.
   *  (default: off)</pre>
   * 
   * <pre> -H
   *  Threshold for cluster sizes.
   *  (default: 0.1)</pre>
   * 
   * <pre> -S &lt;num&gt;
   *  Random number seed.
   *  (default 1)</pre>
   * 
   * <pre> -D
   *  If set, classifier is run in debug mode and
   *  may output additional info to the console</pre>
   * 
   * <pre> -W
   *  Full name of base classifier.
   *  (default: weka.classifiers.rules.ZeroR)</pre>
   * 
   * <pre> 
   * Options specific to classifier weka.classifiers.rules.ZeroR:
   * </pre>
   * 
   * <pre> -D
   *  If set, classifier is run in debug mode and
   *  may output additional info to the console</pre>
   * 
   <!-- options-end -->
   *
   * @param options 	the list of options as an array of strings
   * @throws Exception 	if an option is not supported
   */
  public void setOptions(String[] options) throws Exception {
    String	tmpStr;
    String[] 	tmpOptions;

    tmpStr = Utils.getOption('C', options);
    if (tmpStr.length() != 0)
      setNumClusters(Integer.parseInt(tmpStr));
    else
      setNumClusters(0);

    tmpStr     = Utils.getOption('K', options);
    tmpOptions = Utils.splitOptions(tmpStr);
    if (tmpOptions.length != 0) {
      tmpStr        = tmpOptions[0];
      tmpOptions[0] = "";
      setClusterer((Clusterer) Utils.forName(Clusterer.class, tmpStr, tmpOptions));
    }
    else {
      setClusterer((Clusterer) Utils.forName(Clusterer.class, defaultClustererString(), new String[]{}));
    }

    setEqualWeights(Utils.getFlag('I', options));

    setDistribution(Utils.getFlag('T', options));

    tmpStr = Utils.getOption('H', options);
    if (tmpStr.length() != 0)
      setThreshold(Double.parseDouble(tmpStr));
    else
      setThreshold(0.0);

    super.setOptions(options);
  }

  /**
   * Gets the current settings of the classifier.
   *
   * @return 		an array of strings suitable for passing to setOptions
   */
  public String[] getOptions() {
    Vector<String>	result;
    String[] 		options;
    int			i;
    String		tmpStr;

    result = new Vector<String>();

    result.add("-C"); 
    result.add("" + getNumClusters());
    
    result.add("-K");
    tmpStr = getClusterer().getClass().getName();
    if (getClusterer() instanceof OptionHandler)
      tmpStr += " " + Utils.joinOptions(((OptionHandler) getClusterer()).getOptions());
    tmpStr = tmpStr.trim();
    result.add(tmpStr);

    if (getEqualWeights())
      result.add("-I");

    if (getDistribution())
      result.add("-T");

    result.add("-H");
    result.add("" + getThreshold());
    
    options = super.getOptions();
    for (i = 0; i < options.length; i++)
      result.add(options[i]);
    
    return result.toArray(new String[result.size()]);
  }

  /**
   * String describing default clusterer.
   * 
   * @return 		the default clusterer classname
   */
  protected String defaultClustererString() {
    return FirstKRelativeSquaredEuclideanDistanceClusterer.class.getName();
  }

  /**
   * A silly cludge method that tries to extract centriodds from non centriod 
   * clusterers.
   * 
   * @param data 	the clustered instances
   * @param clusterer 	the clusterer that 
   * @return 		the centriods
   * @throws Exception 	if something was wrong with the clusterer
   */
  public Instances customCentroids(Instances data,Clusterer clusterer) throws Exception{
    // create the centroid array	  
    int numAttas = data.numAttributes();
    int numClusters = clusterer.numberOfClusters();
    double[][] centroidData = new double[numClusters][numAttas];
    double[] weight = new double[numClusters];
    
    // populate the centriod array
    double w0, w1, total;
    double[] point, probs;
    Instance instance;
    if (m_Distribution){
      for (int h = 0; h < data.numInstances(); h++) {
	instance = data.instance(h);
	try {
	  probs = clusterer.distributionForInstance(instance);

	  for (int i = 0; i < probs.length; i++) {
	    if (Utils.eq(probs[i],0))
	      continue;
	    w0 = weight[i];
	    w1 = instance.weight() * probs[i];
	    total = w0+w1;
	    w0 /= total;
	    w1 /= total;
	    point = centroidData[i];
	    for (int j = 0; j <point.length; j++) 
	      point[j] = w0*point[j] + w1 * instance.value(j);

	    weight[i] += instance.weight();
	  }
	}
	catch (Exception e) {
	  // some instances mught not be clustered print stack trace and continue
	  e.printStackTrace();
	}
      }
    }
    else{
      int assignment;
      for (int h = 0; h < data.numInstances(); h++) {
	instance = data.instance(h);
	try {
	  assignment =  clusterer.clusterInstance(instance);


	  w0 = weight[assignment];
	  w1 = instance.weight();
	  total = w0+w1;
	  w0 /= total;
	  w1 /= total;
	  point = centroidData[assignment];
	  for (int i = 0; i <point.length; i++) {
	    point[i] = w0*point[i] + w1 * instance.value(i);
	  }
	  weight[assignment] += instance.weight();
	}
	catch (Exception e) {
	  // some instances mught not be clustered print stack trace and continue
	  e.printStackTrace();
	}
      }

    }
    
    // convert centriods to instances
    Instances clusterCentroids = new Instances(data,numClusters);
    for (int i = 0;i<centroidData.length;i++)
      if (validateInstance(weight[i], centroidData[i]))
	clusterCentroids.add(cluster2instance(weight[i],centroidData[i],data.classIndex()));
    
    return clusterCentroids;
  }  

  /**
   * Validates the weight of a single instance against the threshold.
   * 
   * @param weight	the weight of the instance
   * @param cluster	unused
   * @return		true if passed
   */
  protected boolean validateInstance(double weight, double[] cluster) {
    return (Utils.gr(weight, getThreshold()));
  }

  /**
   * Validate all the instances and return the ones that passed.
   * 
   * @param check	the instances to check
   * @return		the instances that passed
   */
  protected Instances validateInstances(Instances check){
    Instances passed = new Instances(check, 0);
    for (int i = 0; i < check.numInstances(); i++) {
      Instance t = check.instance(i);
      if (Utils.gr(t.weight(), getThreshold()))
	passed.add(t);
    }
    return passed;
  }

  /**
   * Converts a cluster into an instance with weight specified and attribute 
   * values as specified in cluster class index is only used in the case 
   * where the class is not numeric to  ensure that a round value is selected 
   * for the class.
   * 
   * @param cluster 	the attribute values that describe the cluster
   * @param weight 	the weight to be given to the new instance
   * @param classIndex 	the class index to correct non numeric classes
   * @return 		an instance representation of the cluster.
   */
  public Instance cluster2instance(double weight, double[] cluster,  int classIndex) {
    if (!m_numericClass)
      cluster[classIndex] = Math.round(cluster[classIndex]);
    return new Instance(weight,cluster);
  }

  /**
   * Builds clusters for each class with at most k cluster centres remaining 
   * for each class value if there are less than k instances for a class 
   * initially those instances will be returned as the clustered data.
   * 
   * @param data 	the Instances collection from which the clusters are to be built
   * @param k 		the maximum number off instances for each class
   * @return 		the clustered instances as a double array of max length k*numClasses which will be the length in all cases except where there are classes with less than k instances
   * @throws Exception  if something goes wrong
   */
  public Instances clustersForClasses(Instances data, int k) throws Exception {
    Clusterer clusterer = getClusterer();
    int numClasses = data.numClasses();
    Instances clusters = new Instances(data,k);
    m_header = new Instances(data,k);
    
    // declare an array of instances, one Instances collection for each class
    Instances[] perClass = new Instances[numClasses];
    
    // create each Instances collection by copying header info
    for (int i = 0; i < numClasses; i++)
      perClass[i] = new Instances(data, 0);

    // assign each instance to the appropriate Instances set for its class
    for (int i = 0; i < data.numInstances(); i++)
      perClass[(int) data.instance(i).classValue()].add(data.instance(i));

    for (int i = 0; i < numClasses; i++) {
      try {
	perClass[i].setClassIndex(-1);
	clusterer.buildClusterer(perClass[i]);
	perClass[i].setClassIndex(data.classIndex());
      }
      catch (Exception e) {
	e.printStackTrace();
      }
      
      if (clusterer instanceof ClustersRetrievable) {
	clusters = ((ClustersRetrievable)clusterer).getClusterCentroids();
	if (getThreshold() > 0.0)
	  clusters = validateInstances(clusters);
      }
      else{
	clusters = customCentroids(data, clusterer);
      }

      for (int j = 0; j < clusters.numInstances(); j++)
	m_header.add(clusters.instance(j));
    }

    return m_header;
  }

  /**
   * Generate clusters out of the supplied instances, will return an Instances 
   * set ff at most size k iff the class value is numeric, other classes are 
   * split by class before clustering there may be k*numClasses elements in the 
   * resulting Instances set in that case. <p/>
   * 
   * the return value is also stored in m_header, I'm not sure why this seems 
   * to me to be a waste of memory.
   * 
   * @param data 	The set of Instances to be clustered
   * @param k 		the number of clusters to use
   * @return a 		set of clustered instances
   * @throws Exception 	if something goes wrong
   */
  public Instances generateClusters(Instances data, int k) throws Exception {
    if (m_numericClass) {
      Clusterer clusterer = getClusterer();
      int index = data.classIndex();

      try {
	data.setClassIndex(-1);
	clusterer.buildClusterer(data);
      }
      catch (Exception e) {
	e.printStackTrace();
      }
      
      if (clusterer instanceof ClustersRetrievable) {
	m_header = ((ClustersRetrievable)clusterer).getClusterCentroids();
	if (getThreshold() > 0.0)
	  m_header = validateInstances(m_header);
      }
      else{
	m_header = customCentroids(data, clusterer);
      }
      
      m_header.setClassIndex(index);
      
      return m_header;
    }

    return clustersForClasses(data, k);
  }

  /**
   * Builds the classifier.
   *
   * @param instances 	the training data
   * @throws Exception 	if the classifier could not be built successfully
   */
  public void buildClassifier(Instances instances) throws Exception {
    // can classifier handle the data?
    getCapabilities().testWithFail(instances);

    instances = new Instances(instances);
    instances.deleteWithMissingClass();

    // no training data -> 0R model
    if (instances.numInstances() == 0) {
      m_ZeroR = new ZeroR();
      m_ZeroR.buildClassifier(instances);
      return;
    }
    else {
      m_ZeroR = null;
    }
    
    m_missing = new ReplaceMissingValues();
    m_missing.setInputFormat(instances);
    instances = Filter.useFilter(instances, m_missing);

    m_nominaltobinary = new NominalToBinary();
    m_nominaltobinary.setInputFormat(instances);
    instances = Filter.useFilter(instances, m_nominaltobinary);

    m_Ident = new IdenticalInstanceFilter();
    m_Ident.setInputFormat(instances);
    m_Ident.setMerge(true);
    instances = Filter.useFilter(instances, m_Ident);

    m_classIndex = instances.classIndex();
    m_numericClass = instances.attribute(m_classIndex).isNumeric();
    Instance witness = (Instance)instances.instance(0).copy();

    /*
    m_standardize = new Standardize();
    m_standardize.setInputFormat(instances);
    instances = Filter.useFilter(instances, m_standardize);
     */

    m_normalize = new Normalize();
    m_normalize.setInputFormat(instances);
    instances = Filter.useFilter(instances, m_normalize);

    Filter filter = (m_standardize == null) ? m_normalize : m_standardize;

    if (m_numericClass) {
      witness.setValue(m_classIndex, 0);
      filter.input(witness);
      filter.batchFinished();
      Instance res = filter.output();
      m_Blin = res.value(m_classIndex);
      witness.setValue(m_classIndex, 1);
      filter.input(witness);
      filter.batchFinished();
      res = filter.output();
      m_Alin = res.value(m_classIndex) - m_Blin;
    }

    int k = getNumClusters();
    if (k < 1)
      k = (int) Math.ceil(Math.sqrt(instances.numInstances()));

    if (instances.attribute(m_classIndex).isNominal())
      k = (k / instances.attribute(m_classIndex).numValues())+1; // Add one so the mulltiple is greater 

    ((NumberOfClustersRequestable)getClusterer()).setNumClusters(k);
    m_classIndex = instances.classIndex();

    Instances transformed = generateClusters(instances, k);

    if (m_EqualWeights)
      for (int i = 0; i < transformed.numInstances(); i++)
	transformed.instance(i).setWeight(1);

    if (getDebug())
      System.out.println("training with " + transformed.numInstances() + " meta-examples");

    getClassifier().buildClassifier(transformed);

    m_header = new Instances(m_header, 0);
  }

  /**
   * Computes the distribution for a given instance.
   *
   * @param instance the instance for which distribution is computed
   * @return the distribution
   * @throws Exception if the distribution can't be computed successfully
   */
  public double[] distributionForInstance(Instance instance) throws Exception {
    if (m_ZeroR != null)
      return m_ZeroR.distributionForInstance(instance);
    
    m_missing.input(instance);
    m_missing.batchFinished();
    instance = m_missing.output();

    if (getDebug())
      System.out.println(" mis " + instance);

    m_nominaltobinary.input(instance);
    m_nominaltobinary.batchFinished();
    instance = m_nominaltobinary.output();

    if (getDebug())
      System.out.println(" nom " + instance);

    if (m_standardize != null) {
      m_standardize.input(instance);
      m_standardize.batchFinished();
      instance = m_standardize.output();
    }
    else {
      m_normalize.input(instance);
      m_normalize.batchFinished();
      instance = m_normalize.output();
    }

    if (getDebug())
      System.out.println(" sta " + instance);

    double[] result = getClassifier().distributionForInstance(instance);

    if (m_numericClass)
      result[0] = (result[0] - m_Blin) / m_Alin;

    return result;
  }

  /**
   * Returns a description of this classifier as a String.
   *
   * @return a description of this classifier
   */
  public String toString() {
    if (getClassifier() == null)
      return "No classifier built yet!";

    if (m_ZeroR != null)
      return "Couldn't train - using ZeroR model!\n\n" + m_ZeroR.toString();
    
    StringBuffer sb = new StringBuffer();
    sb.append(getClassifierSpec() + "\n\n");
    sb.append(getClassifier().toString());
    
    return sb.toString();
  }

  /**
   * Returns the tip text for this property.
   * 
   * @return 		tip text for this property suitable for
   * 			displaying in the explorer/experimenter gui
   */
  public String numClustersTipText() {
    return "The number of clusters for K-Means to generate.";
  }

  /**
   * Set the number of clusters for K-means to generate.
   *
   * @param value 	the number of clusters to generate.
   */
  public void setNumClusters(int value) {
    if (value >= 0)
      m_numClusters = value;
    else
      System.err.println("At least 1 cluster must be generated, or 0 for sqrt(#examples)!");
  }

  /**
   * Return the number of clusters to generate.
   *
   * @return 		the number of clusters to generate.
   */
  public int getNumClusters() {
    return m_numClusters;
  }

  /**
   * Returns the tip text for this property.
   * 
   * @return 		tip text for this property suitable for
   * 			displaying in the explorer/experimenter gui
   */
  public String clustererTipText() {
    return "The cluster algorithm to use - must implement the NumberOfClustersRequestable interface.";
  }

  /**
   * Sets the clusterer to use.
   * 
   * @param value	the clusterer to use
   */
  public void setClusterer(Clusterer value) {
    if (!(value instanceof NumberOfClustersRequestable))
      throw new IllegalArgumentException(
	    "Clusterer must Implement NumberOfClustersRequestable interface");
    else
      m_Clusterer = value;
  }

  /**
   * Returns the currently set clusterer.
   * 
   * @return		the clusterer in use
   */
  public Clusterer getClusterer(){
    return m_Clusterer;
  }

  /**
   * Returns the tip text for this property.
   * 
   * @return 		tip text for this property suitable for
   * 			displaying in the explorer/experimenter gui
   */
  public String equalWeightsTipText() {
    return "Whether to use equal weights.";
  }

  /**
   * Sets whether to use equal weights or not.
   * 
   * @param value	if true then equal weights are used
   */
  public void setEqualWeights(boolean value) {
    m_EqualWeights = value;
  }

  /**
   * Returns whether equal weights are used.
   * 
   * @return		true if equal weights are used
   */
  public boolean getEqualWeights() {
    return m_EqualWeights;
  }

  /**
   * Returns the tip text for this property.
   * 
   * @return 		tip text for this property suitable for
   * 			displaying in the explorer/experimenter gui
   */
  public String distributionTipText() {
    return "Whether to take the cluster distribution or only the cluster number into account.";
  }

  /**
   * Sets whether to take cluster distribution or only cluster number into 
   * account.
   * 
   * @param value	if true then cluster distribution is used
   */
  public void setDistribution(boolean value) {
    m_Distribution = value;
  }

  /**
   * Returns whether cluster distribution instead of cluster number is taken
   * into account.
   * 
   * @return		true if cluster distribution is used
   */
  public boolean getDistribution() {
    return m_Distribution;
  }

  /**
   * Returns the tip text for this property.
   * 
   * @return 		tip text for this property suitable for
   * 			displaying in the explorer/experimenter gui
   */
  public String thresholdTipText() {
    return "The threshold to use.";
  }

  /**
   * Sets the threshold to use.
   * 
   * @param value	the threshold
   */
  public void setThreshold(double value) {
    if (value >= 0)
      m_Threshold = value;
    else
      System.err.println("Threshold must be at least 0.0!");
  }

  /**
   * Returns the threshold.
   * 
   * @return		the threshold
   */
  public double getThreshold() {
    return m_Threshold;
  }

  /**
   * Returns the revision string.
   * 
   * @return		the revision
   */
  public String getRevision() {
    return RevisionUtils.extract("$Revision: 1.0 $");
  }

  /**
   * Main method for testing this class.
   *
   * @param args 	the options
   */
  public static void main(String[] args) {
    runClassifier(new ClusteredMetaInstanceClassifier(), args);
  }
}
