/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    Prewrapper.java
 *    Copyright (C) 2007 University of Waikato, Hamilton, New Zealand
 *
 */

package weka.classifiers.meta;

import weka.classifiers.RandomizableSingleClassifierEnhancer;
import weka.classifiers.rules.ZeroR;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.RevisionUtils;
import weka.core.TechnicalInformation;
import weka.core.TechnicalInformationHandler;
import weka.core.TechnicalInformation.Field;
import weka.core.TechnicalInformation.Type;
import weka.filters.Filter;
import weka.filters.unsupervised.attribute.NominalToBinary;
import weka.filters.unsupervised.attribute.Normalize;
import weka.filters.unsupervised.attribute.ReplaceMissingValues;
import weka.filters.unsupervised.attribute.Standardize;
import weka.filters.unsupervised.instance.IdenticalInstanceFilter;

/**
 <!-- globalinfo-start -->
 * Applies several filters:<br/>
 * - replacing missing values<br/>
 * - nominal values are turned into binary ones<br/>
 * - identical instances are merged<br/>
 * - final data is normalized<br/>
 * <br/>
 * For more information see:<br/>
 * <br/>
 * Reuben Evans (2008). Clustering for Classification: Using Standard Clustering Methods to Summarise Datasets with Minimal Loss of Classification Accuracy. VDM Verlag.
 * <p/>
 <!-- globalinfo-end -->
 * 
 <!-- technical-bibtex-start -->
 * BibTeX:
 * <pre>
 * &#64;book{Evans2008,
 *    author = {Reuben Evans},
 *    pages = {108},
 *    publisher = {VDM Verlag},
 *    title = {Clustering for Classification: Using Standard Clustering Methods to Summarise Datasets with Minimal Loss of Classification Accuracy},
 *    year = {2008},
 *    ISBN = {3639031636}
 * }
 * </pre>
 * <p/>
 <!-- technical-bibtex-end -->
 *
 <!-- options-start -->
 * Valid options are: <p/>
 * 
 * <pre> -S &lt;num&gt;
 *  Random number seed.
 *  (default 1)</pre>
 * 
 * <pre> -D
 *  If set, classifier is run in debug mode and
 *  may output additional info to the console</pre>
 * 
 * <pre> -W
 *  Full name of base classifier.
 *  (default: weka.classifiers.rules.ZeroR)</pre>
 * 
 * <pre> 
 * Options specific to classifier weka.classifiers.rules.ZeroR:
 * </pre>
 * 
 * <pre> -D
 *  If set, classifier is run in debug mode and
 *  may output additional info to the console</pre>
 * 
 <!-- options-end -->
 * 
 * @author Reuben Evans
 * @version $Revision$
 */
public class Prewrapper
  extends RandomizableSingleClassifierEnhancer
  implements TechnicalInformationHandler {

  /** for serialization. */
  private static final long serialVersionUID = -6557237996772709997L;

  /** for replacing missing values. */
  protected ReplaceMissingValues m_missing;
  
  /** for turning nominal values into binary ones. */
  protected NominalToBinary m_nominaltobinary;
  
  /** the filter to merging the instances.  */
  protected IdenticalInstanceFilter m_Ident;
  
  /** the class index in the dataset. */
  protected int m_classIndex;
  
  /** whether the class is numeric. */
  protected boolean m_numericClass;
  
  /** for normalizing the data. */
  protected Normalize m_normalize;
  
  /** for adjusting numeric classes. */
  protected double m_Blin;
  
  /** for adjusting numeric classes. */
  protected double m_Alin;
  
  /** for standardizing the data. */
  protected Standardize m_standardize;
  
  /** the 0R model, in case we cannot train. */
  protected ZeroR m_ZeroR = null;
  
  /**
   * Returns a string describing classifier.
   * 
   * @return 		a description suitable for
   * 			displaying in the explorer/experimenter gui
   */
  public String globalInfo() {
    return 
        "Applies several filters:\n"
      + "- replacing missing values\n"
      + "- nominal values are turned into binary ones\n"
      + "- identical instances are merged\n"
      + "- final data is normalized\n"
      + "\n"
      + "For more information see:\n\n"
      + getTechnicalInformation().toString();
  }

  /**
   * Returns an instance of a TechnicalInformation object, containing 
   * detailed information about the technical background of this class,
   * e.g., paper reference or book this class is based on.
   * 
   * @return 		the technical information about this class
   */
  public TechnicalInformation getTechnicalInformation() {
    TechnicalInformation 	result;
    
    result = new TechnicalInformation(Type.BOOK);
    result.setValue(Field.AUTHOR, "Reuben Evans");
    result.setValue(Field.YEAR, "2008");
    result.setValue(Field.TITLE, "Clustering for Classification: Using Standard Clustering Methods to Summarise Datasets with Minimal Loss of Classification Accuracy");
    result.setValue(Field.ISBN, "3639031636");
    result.setValue(Field.PAGES, "108");
    result.setValue(Field.PUBLISHER, "VDM Verlag");
    
    return result;
  }

  /**
   * Generates the classifier.
   *
   * @param instances 	set of instances serving as training data 
   * @throws Exception 	if the classifier has not been generated successfully
   */
  public void buildClassifier(Instances instances) throws Exception {
    // can classifier handle the data?
    getCapabilities().testWithFail(instances);
    
    instances = new Instances(instances);
    instances.deleteWithMissingClass();

    // no training data -> 0R model
    if (instances.numInstances() == 0) {
      m_ZeroR = new ZeroR();
      m_ZeroR.buildClassifier(instances);
      return;
    }
    else {
      m_ZeroR = null;
    }

    m_missing = new ReplaceMissingValues();
    m_missing.setInputFormat(instances);
    instances = Filter.useFilter(instances, m_missing);

    m_nominaltobinary = new NominalToBinary();
    m_nominaltobinary.setInputFormat(instances);
    instances = Filter.useFilter(instances, m_nominaltobinary);

    m_Ident = new IdenticalInstanceFilter();
    m_Ident.setInputFormat(instances);
    m_Ident.setMerge(true);
    instances = Filter.useFilter(instances, m_Ident);

    m_classIndex = instances.classIndex();
    m_numericClass = (instances.attribute(m_classIndex).isNumeric());

    /*
	    m_standardize = new Standardize();
	    m_standardize.setInputFormat(instances);
	    instances = Filter.useFilter(instances, m_standardize);
     */

    m_normalize = new Normalize();
    m_normalize.setInputFormat(instances);
    instances = Filter.useFilter(instances, m_normalize);

    Filter filter = (m_standardize == null) ? m_normalize : m_standardize;

    if (m_numericClass) {
      Instance witness = (Instance) instances.instance(0).copy();
      witness.setValue(m_classIndex, 0);
      filter.input(witness);
      filter.batchFinished();
      Instance res = filter.output();
      m_Blin = res.value(m_classIndex);
      witness.setValue(m_classIndex, 1);
      filter.input(witness);
      filter.batchFinished();
      res = filter.output();
      m_Alin = res.value(m_classIndex) - m_Blin;
    }

    getClassifier().buildClassifier(instances);	    
  }
  
  /**
   * Computes the distribution for a given instance.
   *
   * @param instance 	the instance for which distribution is computed
   * @return 		the distribution
   * @throws Exception 	if the distribution can't be computed successfully
   */
  public double[] distributionForInstance(Instance instance) throws Exception {
    if (m_ZeroR != null)
      return m_ZeroR.distributionForInstance(instance);
    
    m_missing.input(instance);
    m_missing.batchFinished();
    instance = m_missing.output();

    if (getDebug())
      System.out.println(" mis " + instance);

    m_nominaltobinary.input(instance);
    m_nominaltobinary.batchFinished();
    instance = m_nominaltobinary.output();

    if (getDebug())
      System.out.println(" nom " + instance);

    if (m_standardize != null) {
      m_standardize.input(instance);
      m_standardize.batchFinished();
      instance = m_standardize.output();
    }
    else {
      m_normalize.input(instance);
      m_normalize.batchFinished();
      instance = m_normalize.output();
    }

    if (getDebug())
      System.out.println(" sta " + instance);

    double[] result = getClassifier().distributionForInstance(instance);

    if (m_numericClass)
      result[0] = (result[0] - m_Blin) / m_Alin;

    return result;
  }
  
  /**
   * Returns the model of the base classifier.
   * 
   * @return		the model string
   */
  public String toString() {
    StringBuffer	result;

    if (m_ZeroR != null)
      return "Couldn't train - using ZeroR model!\n\n" + m_ZeroR.toString();
    
    result = new StringBuffer();
    result.append(getClass().getName().replaceAll(".*\\.", "") + "\n");
    result.append(getClass().getName().replaceAll(".*\\.", "").replaceAll(".", "=") + "\n\n");
    result.append("Model of base classifier: " + getClassifierSpec() + "\n\n");
    result.append(getClassifier().toString());
    
    return result.toString();
  }
  
  /**
   * Returns the revision string.
   * 
   * @return		the revision
   */
  public String getRevision() {
    return RevisionUtils.extract("$Revision: 1.0 $");
  }
  
  /**
   * Main method for testing this class.
   *
   * @param args 	the options
   */
  public static void main(String[] args) {
    runClassifier(new Prewrapper(), args);
  }
}
