/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/*
 *    BisectingKMeans.java
 *    Copyright (C) 2007 University of Waikato, Hamilton, New Zealand
 *
 */

package weka.clusterers;

import weka.core.Attribute;
import weka.core.Capabilities;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Option;
import weka.core.RevisionUtils;
import weka.core.TechnicalInformation;
import weka.core.TechnicalInformationHandler;
import weka.core.Utils;
import weka.core.WeightedInstancesHandler;
import weka.core.Capabilities.Capability;
import weka.core.TechnicalInformation.Field;
import weka.core.TechnicalInformation.Type;
import weka.filters.Filter;
import weka.filters.unsupervised.attribute.ReplaceMissingValues;

import java.util.Enumeration;
import java.util.Vector;

/**
 <!-- globalinfo-start -->
 * Cluster data using the Biscecting k means algorithm
 * <p/>
 <!-- globalinfo-end -->
 * 
 <!-- technical-bibtex-start -->
 * BibTeX:
 * <pre>
 * &#64;book{Evans2008,
 *    author = {Reuben Evans},
 *    pages = {108},
 *    publisher = {VDM Verlag},
 *    title = {Clustering for Classification: Using Standard Clustering Methods to Summarise Datasets with Minimal Loss of Classification Accuracy},
 *    year = {2008},
 *    ISBN = {3639031636}
 * }
 * </pre>
 * <p/>
 <!-- technical-bibtex-end -->
 *
 <!-- options-start -->
 * Valid options are: <p/>
 * 
 * <pre> -N &lt;num&gt;
 *  number of clusters.
 *  (default 2).</pre>
 * 
 * <pre> -S &lt;num&gt;
 *  Random number seed.
 *  (default 10)</pre>
 * 
 <!-- options-end -->
 * 
 * @author Reuben Evans
 * @version $Revision$
 */
public class BisectingKMeans
  extends RandomizableClusterer 
  implements ClustersRetrievable, NumberOfClustersRequestable, 
  	     WeightedInstancesHandler, TechnicalInformationHandler {

  /**
   * The version of this class date and time of last breaking edit .
   */
  private static final long serialVersionUID = 200702271154L;

  /**
   * replace missing values in training instances.
   */
  protected ReplaceMissingValues m_ReplaceMissingFilter;

  /**
   * number of clusters to generate.
   */
  protected int m_NumClusters = 2;

  /**
   * holds the cluster centroids.
   */
  protected Instances m_ClusterCentroids;

  /**
   * Holds the cluster data.
   */
  protected Instances[] m_ClusterMembers;

  /**
   * The number of instances in each cluster.
   */
  protected int[] m_ClusterSizes;

  /**
   * The clusterer.
   */
  protected SimpleKMeans m_Clusterer;

  /**
   * attribute min values.
   */
  protected double[] m_Min;

  /**
   * attribute max values.
   */
  protected double[] m_Max;

  /**
   * default constructor.
   */
  public BisectingKMeans(){
    super();

    m_SeedDefault = 10;
    setSeed(m_SeedDefault);
  }

  /**
   * Returns a string describing this clusterer.
   * 
   * @return 		a description of the evaluator suitable for
   * 			displaying in the explorer/experimenter gui
   */
  public String globalInfo() {
    return "Cluster data using the Biscecting k means algorithm";
  }

  /**
   * Returns an instance of a TechnicalInformation object, containing 
   * detailed information about the technical background of this class,
   * e.g., paper reference or book this class is based on.
   * 
   * @return 		the technical information about this class
   */
  public TechnicalInformation getTechnicalInformation() {
    TechnicalInformation 	result;
    
    result = new TechnicalInformation(Type.BOOK);
    result.setValue(Field.AUTHOR, "Reuben Evans");
    result.setValue(Field.YEAR, "2008");
    result.setValue(Field.TITLE, "Clustering for Classification: Using Standard Clustering Methods to Summarise Datasets with Minimal Loss of Classification Accuracy");
    result.setValue(Field.ISBN, "3639031636");
    result.setValue(Field.PAGES, "108");
    result.setValue(Field.PUBLISHER, "VDM Verlag");
    
    return result;
  }

  /**
   * Returns an enumeration describing the available options.
   *
   * @return 		an enumeration of all the available options.
   */
  public Enumeration listOptions () {
    Vector result = new Vector();

    result.addElement(new Option(
	  "\tnumber of clusters.\n"
	  + "\t(default 2).", 
	  "N", 1, "-N <num>"));

    Enumeration en = super.listOptions();
    while (en.hasMoreElements())
      result.addElement(en.nextElement());

    return  result.elements();
  }

  /**
   * Parses a given list of options. <p/>
   * 
   <!-- options-start -->
   * Valid options are: <p/>
   * 
   * <pre> -N &lt;num&gt;
   *  number of clusters.
   *  (default 2).</pre>
   * 
   * <pre> -S &lt;num&gt;
   *  Random number seed.
   *  (default 10)</pre>
   * 
   <!-- options-end -->
   *
   * @param options 	the list of options as an array of strings
   * @throws Exception 	if an option is not supported
   */
  public void setOptions (String[] options)
  throws Exception {

    String optionString = Utils.getOption('N', options);

    if (optionString.length() != 0) {
      setNumClusters(Integer.parseInt(optionString));
    }

    super.setOptions(options);
  }

  /**
   * Gets the current settings.
   *
   * @return 		an array of strings suitable for passing to setOptions()
   */
  public String[] getOptions () {
    int       	i;
    Vector    	result;
    String[]  	options;

    result = new Vector();

    result.add("-N");
    result.add("" + getNumClusters());

    options = super.getOptions();
    for (i = 0; i < options.length; i++)
      result.add(options[i]);

    return (String[]) result.toArray(new String[result.size()]);	  
  }

  /**
   * Returns default capabilities of the clusterer.
   *
   * @return      the capabilities of this clusterer
   */
  public Capabilities getCapabilities() {
    Capabilities result = super.getCapabilities();

    // attributes
    result.enable(Capability.NOMINAL_ATTRIBUTES);
    result.enable(Capability.NUMERIC_ATTRIBUTES);
    result.enable(Capability.MISSING_VALUES);

    return result;
  }

  /**
   * Generates the clusterer.
   *
   * @param data 	set of instances serving as training data 
   * @throws Exception 	if the clusterer has not been generated successfully
   */
  public void buildClusterer(Instances data) throws Exception {
    // can clusterer handle the data?
    getCapabilities().testWithFail(data);

    m_ClusterCentroids = new Instances(data,m_NumClusters);

    m_ReplaceMissingFilter = new ReplaceMissingValues();
    Instances instances = new Instances(data);
    instances.setClassIndex(-1);
    m_ReplaceMissingFilter.setInputFormat(instances);
    instances = Filter.useFilter(instances, m_ReplaceMissingFilter);

    m_Min = new double [instances.numAttributes()];
    m_Max = new double [instances.numAttributes()];
    for (int i = 0; i < instances.numAttributes(); i++) {
      m_Min[i] = m_Max[i] = Double.NaN;
    }

    for (int i = 0; i < instances.numInstances(); i++) {
      updateMinMax(instances.instance(i));
    }

    // setup the simple K means
    m_Clusterer = new SimpleKMeans();
    // restrict clusters to two since we''re bisecting
    m_Clusterer.setNumClusters(2);
    m_Clusterer.setSeed(m_Seed);

    // setup stats variables
    Instance[] centroids = new Instance[m_NumClusters];
    m_ClusterSizes = new int[m_NumClusters];

    m_ClusterMembers = new Instances[m_NumClusters];
    m_ClusterMembers[0] = new Instances(data);
    
    // pick cluster to split initally data
    int max,i,j;
    m_ClusterSizes[0] = data.numInstances();
    for (i = 1;i<m_NumClusters;i++){
      max = 0;
      for (j = 0;j<i;j++)
	if (m_ClusterSizes[j]>m_ClusterSizes[max])
	  max = j;
      //System.out.println("Max Cluster" + max + " is " + m_ClusterSizes[max]);
      if (m_ClusterSizes[max]<2)
	break;
      m_Clusterer.buildClusterer(m_ClusterMembers[max]);
      if (m_Clusterer.getNumClusters() < 2){
	centroids[max] = m_Clusterer.getClusterCentroids().firstInstance();
	m_ClusterSizes[max] = 1;
	i--;
      }else{
	splitDataset(max,i,centroids);
      }
    }

    // populate centroids
    for (i--; i >= 0; i--)
      m_ClusterCentroids.add(centroids[i]);
    m_ClusterCentroids.compactify();
    m_NumClusters = m_ClusterCentroids.numInstances();
    m_ClusterMembers = null; //tidy up
  }

  /**
   * Splits the dataset.
   * 
   * @param first	the first index in the centroids
   * @param second	the second index in the centroids
   * @param centroids	the centroids to fill
   * @throws Exception	if clusting instances fails
   */
  protected void splitDataset(int first, int second, Instance[] centroids) throws Exception {
    Instances data = m_ClusterMembers[first];
    m_ClusterMembers[first] = new Instances(data,0);
    m_ClusterMembers[second] = new Instances(data,0);
    Instance currentInstance;
    for (int i = 0;i<data.numInstances();i++){
      currentInstance = data.instance(i);
      if (m_Clusterer.clusterInstance(currentInstance) > 0)
	m_ClusterMembers[second].add(currentInstance);
      else
	m_ClusterMembers[first].add(currentInstance);
    }
    
    // tidy up the stats
    Instances temp = m_Clusterer.getClusterCentroids();
    centroids[first]=temp.instance(0);
    centroids[second]=temp.instance(1);
    int[] sizes = m_Clusterer.getClusterSizes();
    m_ClusterSizes[first]= sizes[0];
    m_ClusterSizes[second]=sizes[1];

  }

  /**
   * Classifies a given instance.
   *
   * @param instance 	the instance to be assigned to a cluster
   * @return 		the number of the assigned cluster as an interger
   * 			if the class is enumerated, otherwise the predicted value
   * @throws Exception 	if instance could not be classified successfully
   */
  public int clusterInstance(Instance instance) throws Exception {
    m_ReplaceMissingFilter.input(instance);
    m_ReplaceMissingFilter.batchFinished();
    Instance inst = m_ReplaceMissingFilter.output();

    return clusterProcessedInstance(inst);
  }
  
  /**
   * clusters an instance that has been through the filters.
   *
   * @param instance 	the instance to assign a cluster to
   * @return 		a cluster number
   */
  protected int clusterProcessedInstance(Instance instance) {
    double minDist = Integer.MAX_VALUE;
    int bestCluster = 0;
    for (int i = 0; i < m_NumClusters; i++) {
      double dist = distance(instance, m_ClusterCentroids.instance(i));
      if (dist < minDist) {
	minDist = dist;
	bestCluster = i;
      }
    }
    return bestCluster;
  }

  /**
   * Calculates the distance between two instances.
   *
   * @param first 	the first instance
   * @param second 	the second instance
   * @return 		the distance between the two given instances, between 0 and 1
   */          
  protected double distance(Instance first, Instance second) {  
    double distance = 0;
    int firstI, secondI;

    for (int p1 = 0, p2 = 0; 
    p1 < first.numValues() || p2 < second.numValues();) {
      if (p1 >= first.numValues()) {
	firstI = m_ClusterCentroids.numAttributes();
      } else {
	firstI = first.index(p1); 
      }
      if (p2 >= second.numValues()) {
	secondI = m_ClusterCentroids.numAttributes();
      } else {
	secondI = second.index(p2);
      }
      double diff;
      if (firstI == secondI) {
	diff = difference(firstI, 
	      first.valueSparse(p1),
	      second.valueSparse(p2));
	p1++; p2++;
      } else if (firstI > secondI) {
	diff = difference(secondI, 
	      0, second.valueSparse(p2));
	p2++;
      } else {
	diff = difference(firstI, 
	      first.valueSparse(p1), 0);
	p1++;
      }
      distance += diff * diff;
    }

    return distance;
  }

  /**
   * Computes the difference between two given attribute
   * values.
   * 
   * @param index 	the attribute index
   * @param val1 	the first value
   * @param val2 	the second value
   * @return 		the difference
   */
  protected double difference(int index, double val1, double val2) {
    switch (m_ClusterCentroids.attribute(index).type()) {
      case Attribute.NOMINAL:
	// If attribute is nominal
	if (Instance.isMissingValue(val1) || 
	      Instance.isMissingValue(val2) ||
	      ((int)val1 != (int)val2)) {
	  return 1;
	} else {
	  return 0;
	}
	
      case Attribute.NUMERIC:
	// If attribute is numeric
	if (Instance.isMissingValue(val1) || 
	      Instance.isMissingValue(val2)) {
	  if (Instance.isMissingValue(val1) && 
		Instance.isMissingValue(val2)) {
	    return 1;
	  } else {
	    double diff;
	    if (Instance.isMissingValue(val2)) {
	      diff = norm(val1, index);
	    } else {
	      diff = norm(val2, index);
	    }
	    if (diff < 0.5) {
	      diff = 1.0 - diff;
	    }
	    return diff;
	  }
	} else {
	  return norm(val1, index) - norm(val2, index);
	}
	
      default:
	return 0;
    }
  }

  /**
   * Normalizes a given value of a numeric attribute.
   *
   * @param x 		the value to be normalized
   * @param i 		the attribute's index
   * @return 		the normalized value
   */
  protected double norm(double x, int i) {
    if (Double.isNaN(m_Min[i]) || Utils.eq(m_Max[i],m_Min[i])) {
      return 0;
    }
    else {
      return (x - m_Min[i]) / (m_Max[i] - m_Min[i]);
    }
  }

  /**
   * Updates the minimum and maximum values for all the attributes
   * based on a new instance.
   *
   * @param instance 	the new instance
   */
  protected void updateMinMax(Instance instance) {  
    for (int j = 0;j < m_ClusterCentroids.numAttributes(); j++) {
      if (!instance.isMissing(j)) {
	if (Double.isNaN(m_Min[j])) {
	  m_Min[j] = instance.value(j);
	  m_Max[j] = instance.value(j);
	} else {
	  if (instance.value(j) < m_Min[j]) {
	    m_Min[j] = instance.value(j);
	  } else {
	    if (instance.value(j) > m_Max[j]) {
	      m_Max[j] = instance.value(j);
	    }
	  }
	}
      }
    }
  }

  /**
   * Returns the number of clusters.
   *
   * @return 		the number of clusters generated for a training dataset.
   * @throws Exception 	if number of clusters could not be returned successfully
   */
  public int numberOfClusters() throws Exception {
    return m_NumClusters;
  }

  /**
   * Returns the tip text for this property.
   * 
   * @return 		tip text for this property suitable for
   * 			displaying in the explorer/experimenter gui
   */
  public String numClustersTipText() {
    return "set number of clusters";
  }

  /**
   * set the number of clusters to generate.
   *
   * @param value 		the number of clusters to generate
   * @throws Exception	if number of clusters is negative
   */
  public void setNumClusters(int value) throws Exception {
    if (value <= 0) {
      throw new Exception("Number of clusters must be > 0");
    }
    m_NumClusters = value;
  }

  /**
   * gets the number of clusters to generate.
   *
   * @return 		the number of clusters to generate
   */
  public int getNumClusters() {
    return m_NumClusters;
  }

  /**
   * Gets the the cluster centroids.
   * 
   * @return		the cluster centroids
   */
  public Instances getClusterCentroids() {
    return m_ClusterCentroids;
  }
  
  /**
   * Gets the number of instances in each cluster.
   * 
   * @return		The number of instances in each cluster
   */
  public int[] getClusterSizes() {
    return m_ClusterSizes;
  }
  
  /**
   * return a string describing this clusterer.
   *
   * @return a description of the clusterer as a string
   */
  public String toString() {
    int maxWidth = 0;
    for (int i = 0; i < m_NumClusters; i++) {
      for (int j = 0 ;j < m_ClusterCentroids.numAttributes(); j++) {
	if (m_ClusterCentroids.attribute(j).isNumeric()) {
	  double width = Math.log(Math.abs(m_ClusterCentroids.instance(i).value(j))) /
	  Math.log(10.0);
	  width += 1.0;
	  if ((int)width > maxWidth) {
	    maxWidth = (int)width;
	  }
	}
      }
    }
    StringBuffer temp = new StringBuffer();
    String naString = "N/A";
    for (int i = 0; i < maxWidth+2; i++) {
      naString += " ";
    }
    temp.append("\nBisectingkMeans\n======\n");

    temp.append("\n\nCluster centroids:\n");
    for (int i = 0; i < m_NumClusters; i++) {
      temp.append("\nCluster "+i+"\n\t");
      temp.append("Mean/Mode: ");
      for (int j = 0; j < m_ClusterCentroids.numAttributes(); j++) {
	if (m_ClusterCentroids.attribute(j).isNominal()) {
	  temp.append(" "+m_ClusterCentroids.attribute(j).
		value((int)m_ClusterCentroids.instance(i).value(j)));
	} else {
	  temp.append(" "+Utils.doubleToString(m_ClusterCentroids.instance(i).value(j),
		maxWidth+5, 4));
	}
      }

    }
    temp.append("\n\n");
    return temp.toString();
  }

  /**
   * Returns the revision string.
   * 
   * @return		the revision
   */
  public String getRevision() {
    return RevisionUtils.extract("$Revision: 1.0 $");
  }

  /**
   * Main method for testing this class.
   *
   * @param args 	the options
   */
  public static void main(String[] args) {
    runClusterer(new BisectingKMeans(), args);
  }
}
