/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/*
 *    FirstKRelativeSquaredEuclideanDistanceClusterer.java
 *    Copyright (C) 2007 University of Waikato, Hamilton, New Zealand
 *
 */

package weka.clusterers;

import weka.core.Capabilities;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Option;
import weka.core.RevisionUtils;
import weka.core.TechnicalInformation;
import weka.core.TechnicalInformationHandler;
import weka.core.Utils;
import weka.core.Capabilities.Capability;
import weka.core.TechnicalInformation.Field;
import weka.core.TechnicalInformation.Type;
import weka.filters.unsupervised.attribute.NominalToBinary;
import weka.filters.unsupervised.attribute.ReplaceMissingValues;

import java.util.Enumeration;
import java.util.Random;
import java.util.Vector;

/**
 <!-- globalinfo-start -->
 * Clusterer that implements a radial basis function network.
 * <p/>
 <!-- globalinfo-end -->
 * 
 <!-- technical-bibtex-start -->
 * BibTeX:
 * <pre>
 * &#64;book{Evans2008,
 *    author = {Reuben Evans},
 *    pages = {108},
 *    publisher = {VDM Verlag},
 *    title = {Clustering for Classification: Using Standard Clustering Methods to Summarise Datasets with Minimal Loss of Classification Accuracy},
 *    year = {2008},
 *    ISBN = {3639031636}
 * }
 * </pre>
 * <p/>
 <!-- technical-bibtex-end -->
 *
 <!-- options-start -->
 * Valid options are: <p/>
 * 
 * <pre> -C &lt;number&gt;
 *  Set the number of clusters to generate.
 *  (default = 0, i.e. sqrt(numExamples)).</pre>
 * 
 * <pre> -S &lt;num&gt;
 *  Random number seed.
 *  (default 1)</pre>
 * 
 <!-- options-end -->
 * 
 * @author Reuben Evans
 * @version $Revision$
 */
public class FirstKRelativeSquaredEuclideanDistanceClusterer 
  extends RandomizableClusterer 
  implements NumberOfClustersRequestable, ClustersRetrievable,
             TechnicalInformationHandler {

  /** for serialization. */
  private static final long serialVersionUID = -1207297295522481013L;

  /** Filter used for converting nominal values, because we can't handle them 
   * in the distance function. */
  protected NominalToBinary m_nominaltobinary;

  /** Filter used to replace missing vlaues, because we can't handle them. */
  protected ReplaceMissingValues m_missing;

  /** The number of clusters. */
  protected int m_numClusters = 0;

  /** the clusters created from the instances used internally this and the 
   * weights make up the new instances set. */
  protected double[][] m_clusters;
  
  /** the weights for each cluster. */
  protected double[] m_weight;
  
  /** the header information this also store all the clusters when the 
   * classifier is built. */
  protected Instances m_ClustersCentres;

  /**
   * Returns a string describing this classifier.
   * 
   * @return 		a description of the classifier suitable for
   * 			displaying in the explorer/experimenter gui
   */
  public String globalInfo() {
    return "Clusterer that implements a radial basis function network.";
  }

  /**
   * Returns an enumeration describing the available options.
   *
   * @return 		an enumeration of all the available options
   */
  public Enumeration listOptions() {
    Vector result = new Vector();

    result.addElement(new Option(
	  "\tSet the number of clusters to generate.\n"
	  + "\t(default = 0, i.e. sqrt(numExamples)).",
	  "C", 1, "-C <number>"));

    Enumeration enu = super.listOptions();
    while (enu.hasMoreElements()) {
      result.addElement(enu.nextElement());
    }
    return result.elements();
  }


  /**
   * Parses a given list of options. <p/>
   * 
   <!-- options-start -->
   * Valid options are: <p/>
   * 
   * <pre> -C &lt;number&gt;
   *  Set the number of clusters to generate.
   *  (default = 0, i.e. sqrt(numExamples)).</pre>
   * 
   * <pre> -S &lt;num&gt;
   *  Random number seed.
   *  (default 1)</pre>
   * 
   <!-- options-end -->
   *
   * @param options 	the list of options as an array of strings
   * @throws Exception 	if an option is not supported
   */
  public void setOptions(String[] options) throws Exception {
    String	tmpStr;
    
    tmpStr = Utils.getOption('C', options);
    if (tmpStr.length() != 0)
      setNumClusters(Integer.parseInt(tmpStr));
    else
      setNumClusters(0);

    super.setOptions(options);
  }

  /**
   * Gets the current settings..
   *
   * @return 		an array of strings suitable for passing to setOptions
   */
  public String[] getOptions() {
    int       		i;
    Vector<String>    	result;
    String[]  		options;

    result = new Vector();

    result.add("-C");
    result.add("" + getNumClusters());

    options = super.getOptions();
    for (i = 0; i < options.length; i++)
      result.add(options[i]);

    return result.toArray(new String[result.size()]);	  
  }

  /**
   * Returns an instance of a TechnicalInformation object, containing 
   * detailed information about the technical background of this class,
   * e.g., paper reference or book this class is based on.
   * 
   * @return 		the technical information about this class
   */
  public TechnicalInformation getTechnicalInformation() {
    TechnicalInformation 	result;
    
    result = new TechnicalInformation(Type.BOOK);
    result.setValue(Field.AUTHOR, "Reuben Evans");
    result.setValue(Field.YEAR, "2008");
    result.setValue(Field.TITLE, "Clustering for Classification: Using Standard Clustering Methods to Summarise Datasets with Minimal Loss of Classification Accuracy");
    result.setValue(Field.ISBN, "3639031636");
    result.setValue(Field.PAGES, "108");
    result.setValue(Field.PUBLISHER, "VDM Verlag");
    
    return result;
  }

  /**
   * Returns default capabilities of the clusterer.
   *
   * @return		the capabilities of this clusterer
   */
  public Capabilities getCapabilities() {
    Capabilities result = super.getCapabilities();

    // attributes
    result.enable(Capability.NOMINAL_ATTRIBUTES);
    result.enable(Capability.NUMERIC_ATTRIBUTES);
    result.enable(Capability.DATE_ATTRIBUTES);
    result.enable(Capability.MISSING_VALUES);

    return result;
  }

  /**
   * Generates the clusterer.
   *
   * @param instances 	set of instances serving as training data 
   * @throws Exception 	if the clusterer has not been generated successfully
   */
  public void buildClusterer(Instances instances) throws Exception {
    // can clusterer handle the data?
    getCapabilities().testWithFail(instances);

    instances = new Instances(instances);
    
    int k = getNumClusters();
    if (k < 1)
      k = (int) Math.ceil(Math.sqrt(instances.numInstances()));

    m_ClustersCentres = clustersForData(instances, k);
  }

  /**
   * Predicts the cluster memberships for a given instance.
   *
   * @param instance 	the instance to be assigned a cluster.
   * @return 		an array containing the estimated membership 
   * 			probabilities of the test instance in each cluster (this 
   * 			should sum to at most 1)
   * @throws Exception 	if distribution could not be computed successfully 
   */
  public double[] distributionForInstance(Instance instance) throws Exception {
    int		i;
    double[]	result;
    
    result = new double[numberOfClusters()];
    
    for (i = 0; i < m_ClustersCentres.numInstances(); i++)
      result[i] = relativeSquaredEuclideanDistance(m_ClustersCentres.instance(i), instance.toDoubleArray());

    Utils.normalize(result);
    
    return result;
  }

  /**
   * Returns the number of clusters.
   *
   * @return 		the number of clusters generated for a training dataset.
   * @throws Exception 	if number of clusters could not be returned successfully
   */
  public int numberOfClusters() throws Exception {
    return m_numClusters;
  }

  /**
   * Crreates clusters out of the data, the total number of clusters is given 
   * by k where k is greater than the number of instances each instance is a 
   * cluster when it is less then the data is randomised and mergeWithClosest 
   * is used to merge pairs of instancs into clusters until the number of 
   * clusters is below k.
   * <p/>
   * This is used directly in the numeric case where each Instance has a 
   * different class value and clustering by class would not be useful, in 
   * other cases the instances are clustered by class first before clusters 
   * for data is called.
   * 
   * @param data	the data to use
   * @param k		the number of clusters
   * @return		the cluster centers
   */
  public Instances clustersForData(Instances data, int k) {
    if (k >= data.numInstances()) {
      m_ClustersCentres = data;
      return data;
    }
    m_ClustersCentres = new Instances(data,0);
    double[][] clusters = new double[k][];
    Random r = new Random(getSeed());
    data.randomize(r);
    m_weight = new double[k];
    for(int i = 0; i < k; i++) {
      clusters[i] = data.instance(i).toDoubleArray();
      m_weight[i] = data.instance(i).weight();
    }
    for(int i = k; i < data.numInstances(); i++) {
      mergeWithClosest(data.instance(i),clusters,m_weight);
    }

    for(int i = 0; i < clusters.length; i++) {
      m_ClustersCentres.add(new Instance(m_weight[i],clusters[i]));
    }
    return m_ClustersCentres;
  }	  

  /**
   * Computes the relative distance. <p/>
   * 
   * relative distance: SUM (xi-zi)^2 = SUM xi^2 + SUM zi^2 - 2SUM xi*zi <br/>
   * ==> drop constant SUM xi^2, only iterate over non-zero zi <br/>
   * careful: result can be negative
   * 
   * @param instance	the instance to calculate the relative distance for
   * @param point	the point to use
   * @return		the distance
   */
  public double relativeSquaredEuclideanDistance(Instance instance, double[] point) {
    double sum = 0;

    int n = instance.numValues(); 
    for(int i = 0; i < n; i++) {
      int index = instance.index(i);
      //if (index != m_classIndex) {
      double zi = instance.valueSparse(i);
      sum += zi*zi - 2*zi*point[index];
      //}
    }
    return sum;
  }

  /**
   * For the given instance works out which cluster centre is closest to the instances
   * and merges the instance into the appropriate cluster updating weights accordingly.
   * 
   * @param instance 	the instance to be added
   * @param clusters 	the current set of clusters contains the output as well
   * @param weight 	the current set o wrights and the out put weights after calculation
   */
  public void mergeWithClosest(Instance instance, double[][] clusters, double[] weight) {

    double minD = relativeSquaredEuclideanDistance(instance,clusters[0]);
    int minPoint = 0;
    // for each cluster
    for(int i = 1; i < clusters.length; i++) {
      double d = relativeSquaredEuclideanDistance(instance,clusters[i]);
      if (d < minD) {
	minD = d;
	minPoint = i;
      }
    }
    //... merge ...
    double w0 = weight[minPoint];
    double w1 = instance.weight();
    double total = w0+w1;
    w0 /= total;
    w1 /= total;
    double[] point = clusters[minPoint];
    for(int i = 0; i <point.length; i++) {
      point[i] = w0*point[i] + w1 * instance.value(i);
    }
    weight[minPoint] += instance.weight();
  }
  
  /**
   * Returns the tip text for this property.
   * 
   * @return 		tip text for this property suitable for
   * 			displaying in the explorer/experimenter gui
   */
  public String numClustersTipText() {
    return "The number of clusters to generate.";
  }

  /**
   * Set the number of clusters for K-means to generate.
   *
   * @param value 	the number of clusters to generate.
   */
  public void setNumClusters(int value) {
    if (value > 0) {
      m_numClusters = value;
    }
  }

  /**
   * Return the number of clusters to generate.
   *
   * @return 		the number of clusters to generate.
   */
  public int getNumClusters() {
    return m_numClusters;
  }

  /**
   * Gets the the cluster centroids.
   * 
   * @return		the cluster centroids
   */
  public Instances getClusterCentroids() {
    return m_ClustersCentres;
  }

  /**
   * Returns the revision string.
   * 
   * @return		the revision
   */
  public String getRevision() {
    return RevisionUtils.extract("$Revision: 1.0 $");
  }

  /**
   * Main method for testing this class.
   *
   * @param args 	the options
   */
  public static void main(String[] args) {
    runClusterer(new FirstKRelativeSquaredEuclideanDistanceClusterer(), args);
  }
}
