/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/*
 *    RandomSampler.java
 *    Copyright (C) 2007 University of Waikato, Hamilton, New Zealand
 *
 */

package weka.clusterers;

import weka.core.Instance;
import weka.core.Instances;
import weka.core.RevisionUtils;
import weka.core.TechnicalInformation;
import weka.core.TechnicalInformationHandler;
import weka.core.TechnicalInformation.Field;
import weka.core.TechnicalInformation.Type;

import java.util.Random;

/**
 <!-- globalinfo-start -->
 * A pseudo-clusterer that just uses the input data as centroids, or a randomized sub-set of the input data if there's more instances than clusterers.<br/>
 * <br/>
 * For more information see:<br/>
 * <br/>
 * Reuben Evans (2008). Clustering for Classification: Using Standard Clustering Methods to Summarise Datasets with Minimal Loss of Classification Accuracy. VDM Verlag.
 * <p/>
 <!-- globalinfo-end -->
 * 
 <!-- technical-bibtex-start -->
 * BibTeX:
 * <pre>
 * &#64;book{Evans2008,
 *    author = {Reuben Evans},
 *    pages = {108},
 *    publisher = {VDM Verlag},
 *    title = {Clustering for Classification: Using Standard Clustering Methods to Summarise Datasets with Minimal Loss of Classification Accuracy},
 *    year = {2008},
 *    ISBN = {3639031636}
 * }
 * </pre>
 * <p/>
 <!-- technical-bibtex-end -->
 *
 <!-- options-start -->
 * Valid options are: <p/>
 * 
 * <pre> -S &lt;num&gt;
 *  Random number seed.
 *  (default 1)</pre>
 * 
 <!-- options-end -->
 * 
 * @author Reuben Evans
 * @version $Revision$
 */
public class RandomSampler
  extends RandomizableSingleClustererEnhancer
  implements ClustersRetrievable, NumberOfClustersRequestable,
  	     TechnicalInformationHandler {
  
  /** for serialization. */
  private static final long serialVersionUID = 200704161630L;

  /** the number of clusters. */
  protected int m_NumClusters;

  /** the cluster centroids. */
  protected Instances m_Instances;
  
  /**
   * Returns a string describing classifier.
   * 
   * @return 		a description suitable for
   * 			displaying in the explorer/experimenter gui
   */
  public String globalInfo() {
    return 
        "A pseudo-clusterer that just uses the input data as centroids, or "
      + "a randomized sub-set of the input data if there's more instances "
      + "than clusterers.\n\n"
      + "For more information see:\n\n"
      + getTechnicalInformation().toString();
  }

  /**
   * Returns an instance of a TechnicalInformation object, containing 
   * detailed information about the technical background of this class,
   * e.g., paper reference or book this class is based on.
   * 
   * @return 		the technical information about this class
   */
  public TechnicalInformation getTechnicalInformation() {
    TechnicalInformation 	result;
    
    result = new TechnicalInformation(Type.BOOK);
    result.setValue(Field.AUTHOR, "Reuben Evans");
    result.setValue(Field.YEAR, "2008");
    result.setValue(Field.TITLE, "Clustering for Classification: Using Standard Clustering Methods to Summarise Datasets with Minimal Loss of Classification Accuracy");
    result.setValue(Field.ISBN, "3639031636");
    result.setValue(Field.PAGES, "108");
    result.setValue(Field.PUBLISHER, "VDM Verlag");
    
    return result;
  }

  /**
   * Generates the clusterer.
   *
   * @param data 	set of instances serving as training data 
   * @throws Exception 	if the clusterer has not been generated successfully
   */
  public void buildClusterer(Instances data) throws Exception {
    if (m_NumClusters >= data.numInstances()) {
      m_Instances = data;
      m_NumClusters = data.numInstances();
    }
    else {
      Instances instances = new Instances(data);
      m_Instances = new Instances(data, 0);
      instances.randomize(new Random(getSeed()));
      for (int i = 0; i < m_NumClusters; i++)
	m_Instances.add(instances.instance(i));
    }
  }

  /**
   * Returns the centroids.
   * 
   * @return		the centroids
   */
  public Instances getClusterCentroids() {
    return m_Instances;
  }
  
  /**
   * Returns the tip text for this property.
   * 
   * @return 		tip text for this property suitable for
   * 			displaying in the explorer/experimenter gui
   */
  public String numClustersTipText() {
    return "The number of clusters to generate.";
  }

  /**
   * Returns the number of clusters.
   *
   * @return 		the number of clusters generated for a training dataset.
   */
  public int numberOfClusters(){
    return m_NumClusters;
  }

  /**
   * Sets the number of clusters to generate.
   * 
   * @param value	the number of clusters
   */
  public void setNumClusters(int value) {
    m_NumClusters = value;
  }

  /**
   * Classifies a given instance.
   *
   * @param i	 	the instance to be assigned to a cluster
   * @return 		the number of the assigned cluster as an integer
   * 			if the class is enumerated, otherwise the predicted value
   * @throws Exception 	always throws an exception - not a real clusterer!
   */
  public int clusterInstance(Instance i) throws Exception{
    throw new Exception("This is not a Clusterer!!!");
  }
  
  /**
   * Returns the revision string.
   * 
   * @return		the revision
   */
  public String getRevision() {
    return RevisionUtils.extract("$Revision: 1.0 $");
  }
  
  /**
   * Main method for testing this class.
   *
   * @param args 	use -h for help
   */
  public static void main(String[] args) {
    runClusterer(new RandomSampler(), args);
  }
}
