/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    IdenticalInstanceFilter.java
 *    Copyright (C) 2007 University of Waikato, Hamilton, New Zealand
 */

package weka.filters.unsupervised.instance;

import weka.core.Capabilities;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Option;
import weka.core.RevisionUtils;
import weka.core.Utils;
import weka.core.Capabilities.Capability;
import weka.filters.SimpleBatchFilter;

import java.util.Arrays;
import java.util.Enumeration;
import java.util.Vector;

/**
 <!-- globalinfo-start -->
 * An Instance filter that ether deletes or merges instances that are identical.
 * <p/>
 <!-- globalinfo-end -->
 * 
 <!-- options-start -->
 * Valid options are: <p/>
 * 
 * <pre> -M
 *  Tells the filter to merge rather than delete.</pre>
 * 
 * <pre> -D
 *  Turns on output of debugging information.</pre>
 * 
 <!-- options-end -->
 * 
 * @author Reuben Evans rjee1 at cs dot waikato dot ac dot nz
 * @version $Revision$
 */
public class IdenticalInstanceFilter
  extends SimpleBatchFilter {

  /**
   * The Verion ID date and time of last breaking Edit.
   */
  private static final long serialVersionUID = 200703301210L;

  /**
   * Wether to merge or delete identical instances.
   */
  protected boolean m_Merge = false;

  /**
   * Returns a string describing this filter.
   *
   * @return 		a description of the filter suitable for
   * 			displaying in the explorer/experimenter gui
   */	
  public String globalInfo() {
    return "An Instance filter that ether deletes or merges instances that are identical.";
  }

  /**
   * Returns an enumeration describing the available options.
   *
   * @return 		an enumeration of all the available options.
   */
  public Enumeration listOptions() {
    Vector<Option> result = new Vector<Option>();

    result.addElement(new Option(
	  "\tTells the filter to merge rather than delete.",
	  "M", 0, "-M"));

    Enumeration<Option> e = super.listOptions();
    while(e.hasMoreElements())
      result.add(e.nextElement());

    return result.elements();
  }

  /**
   * Parses the options for this object. <p/>
   *
   <!-- options-start -->
   * Valid options are: <p/>
   * 
   * <pre> -M
   *  Tells the filter to merge rather than delete.</pre>
   * 
   * <pre> -D
   *  Turns on output of debugging information.</pre>
   * 
   <!-- options-end -->
   *
   * @param options 	the list of options as an array of strings
   * @throws Exception 	if an option is not supported
   */
  public void setOptions(String[] options) throws Exception { 
    setMerge(Utils.getFlag('M', options));
    
    super.setOptions(options);
  }

  /**
   * Gets the current settings of the filter.
   *
   * @return an array of strings suitable for passing to setOptions
   */
  public String[] getOptions() {
    Vector<String> result;

    result = new Vector<String>();

    if (getMerge())
      result.add("-M");

    result.addAll(Arrays.asList(super.getOptions()));

    return (String[]) result.toArray(new String[result.size()]);
  }

  /**
   * Sets and returns the output format which is the same as the input format.
   * 
   * @param inputFormat	the input format.
   * @return 		the optput format - an empty copy of the input format
   */
  protected Instances determineOutputFormat(Instances inputFormat) {
    Instances result = new Instances(inputFormat, 0);
    
    return result;
  }

  /** 
   * Returns the Capabilities of this filter. Enables all.
   *
   * @return            the capabilities of this object
   * @see               Capabilities
   */
  public Capabilities getCapabilities() {
    Capabilities 	result;

    result = new Capabilities(this);
    result.setMinimumNumberInstances(0);
    result.enableAllAttributes();
    result.enableAllClasses();
    result.enable(Capability.NO_CLASS);
    
    return result;
  }

  /**
   * Processes the given data (may change the provided dataset) and returns
   * the modified version. This method is called in batchFinished().
   *
   * @param instances   the data to process
   * @return            the modified data
   * @throws Exception  in case the processing goes wrong
   */
  protected Instances process(Instances instances) throws Exception {
    Instances result = new Instances(instances, 0);
    boolean notDuplicate = true;
    for (int i = 0; i < instances.numInstances(); i++){
      for (int j = 0; j < result.numInstances() && notDuplicate; j++) {
	if (equal(instances.instance(i),result.instance(j))){
	  notDuplicate = false;
	  if (m_Merge)
	    result.instance(j).setWeight(result.instance(j).weight()+instances.instance(i).weight());
	}
      }
      
      if (notDuplicate)
	result.add(instances.instance(i));
      
      notDuplicate = true;
    }
    return result;
  }

  /**
   * Checks whether two Instance objects are the same.
   * 
   * @param first	the first instance
   * @param second	the second instance
   * @return		true if equal
   */
  protected boolean equal(Instance first, Instance second){
    if (first.numAttributes() != second.numAttributes())
      return false;
    
    for (int i = 0; i < first.numAttributes(); i++)
      if (first.value(i) != second.value(i))
	return false;
    
    return true;
  }

  /**
   * Returns the tip text for this property.
   * 
   * @return 		tip text for this property suitable for
   * 			displaying in the explorer/experimenter gui
   */
  public String mergeTipText() {
    return "Whether to merge instead of delete identical instances.";
  }

  /**
   * Gets the Value of Merge true if merging false if deleting.
   * 
   * @return true 	if merging instance, false if deleting
   */
  public boolean getMerge() {
    return m_Merge;
  }

  /**
   * Sets the merge option.
   * 
   * @param value 	wether the filter should merge(true) or delete(false)
   */
  public void setMerge(boolean value) {
    this.m_Merge = value;
  }	
  
  /**
   * Returns the revision string.
   * 
   * @return		the revision
   */
  public String getRevision() {
    return RevisionUtils.extract("$Revision: 1.0 $");
  }

  /**
   * Main method for testing this class.
   *
   * @param args 	should contain arguments to the filter: use -h for help
   */
  public static void main(String[] args) {
    runFilter(new IdenticalInstanceFilter(), args);
  }
}
